#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <signal.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <linux/kd.h>
#include <linux/keyboard.h>
#include "getfd.h"
#include "nls.h"
#include "version.h"
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/extensions/XTest.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <sys/types.h>

int fd;
int oldkbmode;
struct termios old;

Display *display;
unsigned int keycode;
int debug = 0;			/* debug flag */
int daemonize = 1;		/* daemonize by default */
int force = 0;
int killflag = 0;

#define PID_FILE "/var/run/rawkeybind.pid"

/*
 * version 0.81 of showkey would restore kbmode unconditially to XLATE,
 * thus making the console unusable when it was called under X.
 */
static void get_mode(void)
{
	char *m;

	if (ioctl(fd, KDGKBMODE, &oldkbmode)) {
		perror("KDGKBMODE");
		exit(1);
	}
	switch (oldkbmode) {
	case K_RAW:
		m = "RAW";
		break;
	case K_XLATE:
		m = "XLATE";
		break;
	case K_MEDIUMRAW:
		m = "MEDIUMRAW";
		break;
	case K_UNICODE:
		m = "UNICODE";
		break;
	default:
		m = _("?UNKNOWN?");
		break;
	}
	if (debug)
		printf(_("kb mode was %s\n"), m);
	/*
	   no need for this, we alredy run in X11 only:
	   if (oldkbmode != K_XLATE) {
	   printf(_("[ if you are trying this under X, it might not work\n"
	   "since the X server is also reading /dev/console ]\n"));
	   }
	   printf("\n");
	 */
}

static void clean_up(void)
{
	if (ioctl(fd, KDSKBMODE, oldkbmode)) {
		perror("KDSKBMODE");
		exit(1);
	}
	if (tcsetattr(fd, 0, &old) == -1)
		perror("tcsetattr");
	unlink(PID_FILE);
	close(fd);
}

static void die(int x)
{
	if (debug)
		printf(_("caught signal %d, cleaning up...\n"), x);
	clean_up();
	exit(1);
}

static void watch_dog(int x)
{
	clean_up();
	exit(0);
}

static int srand_called = 0;

char *random_string(char *dest)
{
	size_t len = 0;
	char *p = dest;
	int three_in_a_row = 0;
	int arr[128] = { 0x0 };

	/* be sure to have called srand exactly one time */
	if (!srand_called) {
		srandom(time(NULL) + getpid());
		srand_called = 1;
	}
	*dest = 0x0;		/* int the destination string */
	for (len = 8; len; len--, p++) {	/* gen characters */
		char *q = dest;
		*p = (rand() % 2) ? rand() % 26 + 97 : rand() % 10 + 48;
		p[1] = 0x0;
		arr[*p]++;	/* check values */
		if (arr[*p] == 3) {
			for (q = dest; q[2] > 0 && !three_in_a_row; q++)
				if (*q == q[1] && q[1] == q[2])
					three_in_a_row = 1;
		}
		if (three_in_a_row || arr[*p] > 3)
			return random_string(dest);	/* values do not pass try again */
	}
	return dest;
}

static void usage(void)
{
	fprintf(stderr, _("rawkeybind version %s\n\n"
			  "usage: rawkeybind [options...]\n"
			  "\n"
			  "valid options are:\n"
			  "\n"
			  "	-h --help	display this help text\n"
			  "	-d --debug	display debug output\n"
			  "	-f --force	force stale lock and start or restart process\n"
			  "	-k --kill	shutdown rawkeybind\n"
			  "	-n --nodetach	do not detach from stdin\n"
			  "	-v --version	display version\n"),
		PACKAGE_VERSION);
	exit(1);
}

int main(int argc, char *argv[])
{
	const char *short_opts = "hdvfnk";
	const struct option long_opts[] = {
		{"help", no_argument, NULL, 'h'},
		{"debug", no_argument, NULL, 'd'},
		{"version", no_argument, NULL, 'v'},
		{"nodetach", no_argument, NULL, 'n'},
		{"force", no_argument, NULL, 'f'},
		{"kill", no_argument, NULL, 'k'},
		{NULL, 0, NULL, 0}
	};
	int c, status;
	int show_keycodes = 1;
	int print_ascii = 0;
	/* config file is hardcoded -- this program is suid */
	FILE *cfg;
	char line[1024] = { 0x0 };
	int lcount = 0, ll = 0;
	/* translation table */
	long kfrom[1024];
	long kto[1024];
	pid_t pid, sid;
	FILE *pidfile;
	char r[11];

	struct termios new;
	unsigned char buf[18];	/* divisible by 3 */
	int i, n;

	/* make a random string for PS so we can ensure that pid 
	   from /var/run/rawkeybind.pid  really belongs to us */
	random_string(r);
	r[10] = '\0';

	/* run setuid to read /dev/console under X11 */
	status = seteuid(0);
	if (status < 0) {
		fprintf(stderr, "Couldn't set uid.\n");
		exit(status);
	}
	if (debug)
		printf("Setuid now\n");

	set_progname(argv[0]);

	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE_NAME, LOCALEDIR);
	textdomain(PACKAGE_NAME);

	while ((c = getopt_long(argc, argv, short_opts, long_opts, NULL)) != -1) {
		switch (c) {
		case 'd':
			debug = 1;
			printf(_("Running in debug mode\n"));
			break;
		case 'n':
			daemonize = 0;
			break;
		case 'f':
			force = 1;
			break;
		case 'k':
			killflag = 1;
			break;
		case 'v':
			print_version_and_exit();
		case 'h':
		case '?':
			usage();
		}
	}

	if (optind < argc)
		usage();

	/* check for pid file existance */
	pidfile = fopen(PID_FILE, "r");
	if (pidfile != NULL) {
		if (force || killflag) {
			char buf1[1024], buf2[1024];
			int p;
			FILE *cmdline;
			char cmdbuf[1024];
			if (debug)
				printf("Forcing the lock file\n");
			if (fgets(buf1, sizeof(buf1), pidfile) != NULL
			    && fgets(buf2, sizeof(buf2), pidfile) != NULL) {
				p = atoi(buf1);
				sprintf(cmdbuf, "/proc/%d/cmdline", p);
				cmdline = fopen(cmdbuf, "r");
				if (cmdline != NULL) {
					if (fgets
					    (cmdbuf, sizeof(cmdbuf),
					     cmdline) == NULL) {
						printf
						    ("Failure to read /proc/%d/cmdline\n",
						     p);
						exit(EXIT_FAILURE);
					}
					if (!strncmp(cmdbuf, buf2, 19)) {
						fclose(pidfile);
						unlink(PID_FILE);
						kill(p, SIGALRM);
						printf("Shutdown process %d\n",p);
						if (killflag)
							exit(EXIT_SUCCESS);
					} else {
						printf ("Can't kill pid %d because it's command line:\n\n%s\n\ndoes not match\
								rawkeybind process title from pid file /var/run/rawkeybind.pid: \n\n%s\n",
						     p, cmdbuf, buf2);
						exit(EXIT_FAILURE);
					}
				}
			}
		} else {
			printf
			    ("Pid file %s alredy exists. Use -f option to force the lock.",
			     PID_FILE);
			exit(EXIT_FAILURE);
		}
	} else {
		/* check for kill flag */
		if (killflag) 
			exit(EXIT_SUCCESS);
	}

	/* change program title to rawkeybind <randomstring> */
	sprintf(argv[0], "rawkeybind %s", r);

	display = XOpenDisplay(NULL);
	if (!display) {
		fprintf(stderr, "Can't open display.\n");
		exit(2);
	}

	/* parse config */
	cfg = fopen("/etc/rawkeybind.conf", "r");
	if (cfg == NULL) {
		perror("Can't open /etc/rawkeybind.conf");
		exit(EXIT_FAILURE);
	}
	while (fgets(line, sizeof(line), cfg) != NULL) {
		char *p, *endptr;
		int val, j;
		char tmpfrom[1024], tmpto[1024];
		ll++;
		if (strlen(line) > 0 && line[strlen(line) - 1] == '\n')
			line[strlen(line) - 1] = '\0';
		if (debug)
			printf("Parsing line '%s'\n", line);
		if (strlen(line) == 0)
			continue;
		for (j = strlen(line); j > 0; j--) {
			if (line[j] == ' ')
				line[j] = '\0';
		}
		if (line[0] == '#')
			continue;
		/* kfrom */
		p = strtok(line, ":");
		strcpy(tmpfrom, p);
		errno = 0;
		val = strtol(tmpfrom, &endptr, 16);
		if ((errno == ERANGE && (val == LONG_MAX || val == LONG_MIN))
		    || (errno != 0 && val == 0)) {
			perror("strtol");
			exit(EXIT_FAILURE);
		}
		if (endptr == tmpfrom) {
			fprintf(stderr, "No digits were found, line %d\n", ll);
			exit(EXIT_FAILURE);
		}
		kfrom[lcount] = val;
		/* kto */
		p = strtok('\0', ":");
		strcpy(tmpto, p);
		errno = 0;
		val = strtol(tmpto, &endptr, 16);
		if ((errno == ERANGE && (val == LONG_MAX || val == LONG_MIN))
		    || (errno != 0 && val == 0)) {
			perror("strtol");
			exit(EXIT_FAILURE);
		}
		if (endptr == tmpto) {
			fprintf(stderr, "No digits were found, line %d\n", ll);
			exit(EXIT_FAILURE);
		}
		kto[lcount] = val;

		if (debug)
			printf("Will fake 0x%lx when 0x%lx is pressed\n",
			       kto[lcount], kfrom[lcount]);
		lcount++;
	}
	if (debug)
		printf("Total %d lines\n", lcount);
	fclose(cfg);

	fd = getfd(NULL);

	/* the program terminates when there is no input for 10 secs */
	signal(SIGALRM, watch_dog);

	/*
	   if we receive a signal, we want to exit nicely, in
	   order not to leave the keyboard in an unusable mode
	 */
	signal(SIGHUP, die);
	signal(SIGINT, die);
	signal(SIGQUIT, die);
	signal(SIGILL, die);
	signal(SIGTRAP, die);
	signal(SIGABRT, die);
	signal(SIGIOT, die);
	signal(SIGFPE, die);
	signal(SIGKILL, die);
	signal(SIGUSR1, die);
	signal(SIGSEGV, die);
	signal(SIGUSR2, die);
	signal(SIGPIPE, die);
	signal(SIGTERM, die);
#ifdef SIGSTKFLT
	signal(SIGSTKFLT, die);
#endif
	signal(SIGCHLD, die);
	signal(SIGCONT, die);
	signal(SIGSTOP, die);
	signal(SIGTSTP, die);
	signal(SIGTTIN, die);
	signal(SIGTTOU, die);

	get_mode();
	if (tcgetattr(fd, &old) == -1)
		perror("tcgetattr");
	if (tcgetattr(fd, &new) == -1)
		perror("tcgetattr");

	new.c_lflag &= ~(ICANON | ECHO | ISIG);
	new.c_iflag = 0;
	new.c_cc[VMIN] = sizeof(buf);
	new.c_cc[VTIME] = 1;	/* 0.1 sec intercharacter timeout */

	if (tcsetattr(fd, TCSAFLUSH, &new) == -1)
		perror("tcsetattr");
	if (ioctl(fd, KDSKBMODE, show_keycodes ? K_MEDIUMRAW : K_RAW)) {
		perror("KDSKBMODE");
		exit(1);
	}

	/* daemonize */

	if (daemonize) {
		/* Fork off the parent process */
		pid = fork();
		if (pid < 0) {
			exit(EXIT_FAILURE);
		}
		/* If we got a good PID, then
		   we can exit the parent process. */
		if (pid > 0) {
			exit(EXIT_SUCCESS);
		}

		/* Change the file mode mask */
		umask(0);

		/* Create a new SID for the child process */
		sid = setsid();
		if (sid < 0) {
			/* Log the failure */
			exit(EXIT_FAILURE);
		}

		/* Change the current working directory */
		if ((chdir("/")) < 0) {
			/* Log the failure */
			exit(EXIT_FAILURE);
		}

		/* save pid file */
		pidfile = fopen(PID_FILE, "w");
		if (pidfile == NULL) {
			perror("can't open pidfile");
			exit(EXIT_FAILURE);
		}
		fprintf(pidfile, "%d\n", getpid());
		fprintf(pidfile, "rawkeybind %s\n", r);
		fclose(pidfile);

		/* Close out the standard file descriptors */
		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		close(STDERR_FILENO);
	}

	/* infinite loop */
	while (1) {
		n = read(fd, buf, sizeof(buf));
		i = 0;
		while (i < n) {
			int kc, j;
			char *s;

			s = (buf[i] & 0x80) ? _("release") : _("press");

			if (i + 2 < n && (buf[i] & 0x7f) == 0
			    && (buf[i + 1] & 0x80) != 0
			    && (buf[i + 2] & 0x80) != 0) {
				kc = ((buf[i + 1] & 0x7f) << 7) |
				    (buf[i + 2] & 0x7f);
				i += 3;
			} else {
				kc = (buf[i] & 0x7f);
				i++;
			}
			if (debug)
				printf(_("keycode 0x%x %s\n"), kc, s);
			for (j = 0; j < lcount; j++) {
				if (kc == kfrom[j]) {
					if (!strcmp(s, "press")) {
						if (debug)
							printf(_
							       ("faking 0x%lx press\n"),
							       kto[j]);
						keycode =
						    XKeysymToKeycode(display,
								     kto[j]);
						XTestFakeKeyEvent(display,
								  keycode, True,
								  0);
					} else {
						if (debug)
							printf(_
							       ("faking 0x%lx release\n"),
							       kto[j]);
						keycode =
						    XKeysymToKeycode(display,
								     kto[j]);
						XTestFakeKeyEvent(display,
								  keycode,
								  False, 0);
					}
					XFlush(display);
				}
			}

		}
	}

	clean_up();
	exit(0);
}
